"""
xcover 
======

An exact cover with colors package.

The main function provided by this packages is "covers" which returns
a generator yielding solutions to exact cover problems.

There are currently four solvers:
    covers -- exact cover with colors
    cover_zdd -- exact cover with colors, but yields nodes of a ZDD not solutions
    covers_bool -- exact cover for problem specified as an array of bool
    covers_bool_zdd -- as above, but yields nodes of a ZDD not solutions
    
The underlying numba-accelerated algorithms behind these solvers can be called directly using either
    algorithm_c -- exact cover with colors
    algorithm_z -- exact cover with colors, but yields nodes of a ZDD
but these require the problems to be first cast in sparse matrix form.
"""

__version__ = "0.2.5"
__author__ = "John F. Rudge"

from .solvers import covers, covers_bool, covers_zdd, covers_bool_zdd
from .dancing_cells import algorithm_c
from .dancing_cells_zdd import algorithm_z



"""Simple file i/o"""

from .solvers import items


def read_xcover_from_file(filename):
    """
    Simple parser for files decribing exact cover problems in Knuth's format
    """
    with open(filename, "r", encoding="utf-8") as cover_file:
        first = True
        options = []
        for line in cover_file.readlines():
            if first:
                if (
                    line[0] == "|" or line[0] == "/" or line == "\n"
                ):  # a comment line, ignore
                    continue
                first = False
                split_pri_sec = line.rstrip("\n").lstrip().split("|")
                primary = split_pri_sec[0].split()
                if len(split_pri_sec) > 1:
                    secondary = split_pri_sec[1].split()
                    colored = True
                else:
                    secondary = None
                    colored = False
            else:
                splits = line.rstrip("\n").lstrip().split()
                options.append(splits)

    return options, primary, secondary, colored


def write_xcover_to_file(
    filename, options, primary=None, secondary=None, colored=False, item_to_string=None
):
    """
    Write options to knuth-formatted DLX file.

    item_to_string is a function that describes the mapping from items to
    strings in the file. If None, a simple mapping which removes spaces is used.
    """

    # A single list of all the options
    all_opts = [o for opt in options for o in opt]

    # Work out explicit primary and secondary item lists
    primary, secondary = items(all_opts, primary, secondary, colored)

    if item_to_string is None:

        def item_to_string(item):
            return str(item).replace(" ", "")

    with open(filename, "w", encoding="utf-8") as cover_file:
        for item in primary:
            cover_file.write(item_to_string(item))
            cover_file.write(" ")
        if secondary:
            cover_file.write("| ")
            for item in secondary:
                cover_file.write(item + " ")
        cover_file.write("\n")
        for option in options:
            for item in option:
                cover_file.write(item_to_string(item))
                cover_file.write(" ")
            cover_file.write("\n")
